/*
 * @(#) RFIDReader.java
 * 
 * RFID Tag Reader API
 * COMP150-TUI: Tangible User Interfaces
 * Tufts University, Department of Computer Science
 *
 */
package rfid;

import java.io.*;
import javax.comm.*;

import java.util.List;
import java.util.Arrays;

/**
 * Interface for an RFID tag reader.
 *
 * @author apokro01@cs.tufts.edu
 * @author michael.horn@tufts.edu
 * @author benjamin.gemmill@tufts.edu
 * @version $Revision$, $Date$
 */
public class RFIDReader implements Runnable {


   /** Name of reader's serial port (e.g. "COM1") */ 
   protected String port;

   /** Serial port input stream */   
   protected InputStream cin;
   
   /** Serial port output stream */
   protected OutputStream cout;

   /** RFIDListener list */
   protected List listeners;

   /** For receiving data from the reader */
   private byte[] readBuffer;

   /** All tags from the most recent inventory */
   protected RFIDTag [] lastInventory;


   /** Flag used to stop polling thread */
   protected boolean stopped;
   
/**
 * Constructs an RFID reader with the name of a serial port
 * (e.g. "COM1"), and opens connection to the reader. 
 */
   public RFIDReader(String port) throws RFIDException {
      try {
         SerialPort sPort = openPort(port);
         
         this.port          = port;
         this.listeners     = new java.util.ArrayList();
         this.lastInventory = new RFIDTag[0];
         this.readBuffer    = new byte[1024];
         this.cin           = sPort.getInputStream();
         this.cout          = sPort.getOutputStream();
	 this.stopped       = false;
         new Thread(this).start();
      }
      catch (Exception x) {
         throw new RFIDException(x);
      }
   }

   
/**
 * Open and configure a SerialPort connection to the reader
 */
   protected SerialPort openPort(String port) throws Exception {
      CommPortIdentifier portId =
      CommPortIdentifier.getPortIdentifier(port);
      SerialPort sPort = (SerialPort)portId.open("JavaRFID",2000);
      sPort.setSerialPortParams(38400,
                                SerialPort.DATABITS_8,
                                SerialPort.STOPBITS_1,
                                SerialPort.PARITY_EVEN);
      sPort.enableReceiveTimeout(6000);  // 6 second timeout
      return sPort;
   }

   
/**
 * Add a listener to be notified of RFID events.
 */
   public void addRFIDListener(RFIDListener listener) { 
       this.listeners.add(listener);
   }

/**
 * Tells the reader to stop polling the reader in a separate
 * thread. 
 */
    public void stopScanner() {
	this.stopped = true;
    }    
   
/**
 * Takes an inventory of all tags within range of the reader.
 * Returned tags will have their RFID and DSFID numbers set.
 */
   public synchronized RFIDTag [] inventory() throws RFIDException {
      try {

         /*
          * Prepare instruction
          */
         byte [] packet = new byte[7];
         packet[0] = 7;            // number of bytes in packet
         packet[1] = (byte)0xFF;   // com address (always 255)
         packet[2] = (byte)0xB0;   // instruction type code
         packet[3] = 0x01;         // inventory code
         packet[4] = 0x00;         // new inventory mode
         
         // compute crc-16 checksum
         int crc = crc16(packet, 5);
         packet[5] = (byte)(crc & 0xFF);
         packet[6] = (byte)((crc >> 8) & 0xFF);
         
         
         /*
          * Send instruction
          */
         printAsHex(packet);
         this.cout.write(packet);
         this.cout.flush();
         
         
         /*
          * Get response
          */
         int count = this.cin.read(readBuffer);
         if (count <= 0) {
            throw new RFIDException("No response from reader.");
         }
         printAsHex(readBuffer, count);
         
         
         RFIDTag[] thisInventory;
         if(readBuffer[3] != 0) {
            thisInventory = new RFIDTag[0];
         } else {
            thisInventory = new RFIDTag[readBuffer[4]];
         }
         
         
         /*
          * note: each returned tag contains 10 bytess:
          * 1 byte transponder type
          * 1 byte dsfid
          * 8 bytes rfid
          */
         for(int i = 0; i < thisInventory.length; i++)
         {
            long rfid = 0;
            for (int j=0; j<8; j++) {              	   
               rfid |=
               ((readBuffer[7+j+10*i] & 0x00FFL) << ((7-j)*8));
            }
            
            thisInventory[i] = new RFIDTag(this, rfid);
            thisInventory[i].setDSFID(readBuffer[6+9*i]);
         }
         
         fireEvents(thisInventory);
         
         return thisInventory;
      }
      catch (IOException iox) {
         throw new RFIDException(iox);
      }
   }

/**
 * Sends a reset to the reader and checks the returned
 * packet to make sure that the reader is hooked up 
 * correctly. 
 */
    public synchronized void hello() throws RFIDException {
	try {
	    byte [] packet = new byte[7];
	    packet[0] = 7;           // number of bytes in packet
	    packet[1] = (byte)0xFF;  // com address (always 255?)
	    packet[2] = (byte)0xB0;  // instruction type code
	    packet[3] = 0x26;        // reset code
	    packet[4] = 0;

	    // compute crc-16 checksum
	    int crc = crc16(packet, 5);
	    packet[5] = (byte)(crc & 0xFF);
	    packet[6] = (byte)((crc >> 8) & 0xFF);

	    printAsHex(packet);
	    this.cout.write(packet);
	    this.cout.flush();
	    
	    // Response buffer should be:
	    //   0x06  0x00  0xB0   0x00  0xD5  0x72
	    int count = this.cin.read(readBuffer);
	    printAsHex(readBuffer, count);
	    
	    if (count != 6 ||
		readBuffer[0] != (byte)0x06 ||
		readBuffer[1] != (byte)0x00 ||
		readBuffer[2] != (byte)0xB0 ||
		readBuffer[3] != (byte)0x00 ||
		readBuffer[4] != (byte)0xD5 ||
		readBuffer[5] != (byte)0x72) {
		throw new RFIDException("Invalid reader response.");
	    }
	}
	catch (IOException iox) {
	    throw new RFIDException(iox);
	}
    }
    

/**
 * Resets the reader to its ready state.
 */
   public synchronized void reset() throws RFIDException {
      try {
         byte [] packet = new byte[7];
         packet[0] = 7;           // number of bytes in packet
         packet[1] = (byte)0xFF;  // com address (always 255?)
         packet[2] = (byte)0xB0;  // instruction type code
         packet[3] = 0x26;        // reset code
         packet[4] = 0;

         // compute crc-16 checksum
         int crc = crc16(packet, 5);
         packet[5] = (byte)(crc & 0xFF);
         packet[6] = (byte)((crc >> 8) & 0xFF);

         printAsHex(packet);
         this.cout.write(packet);
         this.cout.flush();

         int count = this.cin.read(readBuffer);
         if (count > 0) {
            printAsHex(readBuffer, count);
         }
      }
      catch (IOException iox) {
         throw new RFIDException(iox);
      }
   }

   

/**
 * Write a new DSFID number to a tag
 */
   protected synchronized void writeDSFID(RFIDTag tag)
      throws RFIDException {
      
      byte[] data = new byte[16];
      
      data[0] = (byte) 0x10;    //length of data. 
      data[1] = (byte) 0xFF;    //"Data Bus Address", always 255
      data[2] = (byte) 0xB0;    //ISO Host OpCode
      data[3] = (byte) 0x29;    //OpCode
      data[4] = (byte) 0x01;    //addressed mode (writing to specific tag)
      
      //bytes 5 through 12 house the unique ID of the tag (the RFID)
      byte currByte;
      int shiftCount;
      
      for(int x = 0; x < 8; x++) {           
         shiftCount = (7 - x) * 8;   //shift that many places right
         currByte = (byte) ((tag.rfid >> shiftCount) & 0x000000FF);
         data[5+x] = currByte;    
      }
      
      data[13] = (byte) tag.dsfid; //the actual data to write
      
      //CRC
      int crc = crc16(data, 14);
      data[14] = (byte)(crc & 0xFF);      
      data[15] = (byte)((crc >> 8) & 0xFF);
      
      try {
         printAsHex(data);           
         
         //begin actual read/write operations. 
         cout.write(data);
         cout.flush();
         
         int count = cin.read(readBuffer);
         if (count > 0) {
            printAsHex(readBuffer, count);
         }
         
         //bomb out if tag isn't there
         if(readBuffer[3] != 0) {
            throw new RFIDException("Tag not present: " + tag.rfid);
         }
      }
      catch(IOException e) {
         throw(new RFIDException(e));
      }    
   }
   
   
/**
 * Write up to 256 bytes of data onto an RFID tag. 
 */
   protected synchronized void writeData(RFIDTag tag)
      throws RFIDException {
      byte[] data = new byte[146];
      int startBlock = 0, x;
      
      //5 bytes here
      //length of data will be set later, since it can vary
      data[0] = (byte) 146;
      data[1] = (byte) 0xFF;     //stupid com-adr thing
      data[2] = (byte) 0xB0;     //ISO Host OpCode
      data[3] = (byte) 0x24;     //"Write Multiple Blocks" opcode
      data[4] = (byte) 0x01;     //addressed mode
      
      //8 bytes here
      //bytes 5 through 12 house the unique ID of the tag (the RFID).      
      byte currByte;
      int shiftCount;
      for(x = 0; x < 8; x++) {
         shiftCount = (7 - x) * 8;            //shift that many places right          
         currByte = (byte) ((tag.rfid >> shiftCount) & 0x000000FF);   //apply bitmask
         data[5+x] = currByte;       
      }
      
      //3 + 0 to 128 bytes here
      while(startBlock < 64) {
         //starting address to write to
         data[13] = (byte) startBlock;
         
         //number of blocks to write
         data[14] = (byte) 32;
         
         //block size, depends on rfid tag. Equal to 128 / # blocks.
         data[15] = (byte) 0x04; 
         
         //actual data blocks, break out if we go past the end of the string
         for(x = 0; x < 128; x++) {
            if((startBlock * 4) + x >= tag.data.length()) {
               //pad with zeroes
               data[x+16] = (byte) 0x00;
            }
               else {
                  // start writing data at the 16th-byte
                  data[x + 16] = (byte) tag.data.charAt((startBlock*4) + x);
               }
         }           
         
         //CRC for 2 additional bytes
         int crc = crc16(data, x + 16); //30 D6
         data[x + 16] = (byte)(crc & 0xFF);      
         data[x + 17] = (byte)((crc >> 8) & 0xFF);
         
         try {
            //System.out.println();
            //System.out.println("Writing DATA:");
            //System.out.print(">>> ");
            printAsHex(data);           
            
            //begin actual read/write operations. 
            cout.write(data); //5 + 8 + 3 + 2 + x where x is the number of actual data characters
            cout.flush();
            
            int count = cin.read(readBuffer);
            if (count > 0) {
               //System.out.print("<<< ");
               printAsHex(readBuffer, count);
            }
                //System.out.println();   
            
                //bomb out if tag isn't there
            if(readBuffer[3] != 0)
               throw new RFIDException("Tag not present: " + tag.rfid);
         }
         catch(IOException e) {
            throw(new RFIDException(e));
         }          
         startBlock += 32;
      }
   }
   
   
/**
 * Read the entire 256 bytes of data from an RFID tag
 * Note that this will completely erase the data within the tag object 
 */
   protected synchronized void readData(RFIDTag tag)
      throws RFIDException {
       byte output[] = new byte[2000];
       int x, charIndex = 0, c = 0;
       
       /* Prepare instruction, 5 bytes */
       byte [] packet = new byte[17];
       packet[0] = 17;            // number of bytes in packet
       packet[1] = (byte)0xFF;   // com address (always 255)
       packet[2] = (byte)0xB0;   // instruction type code
       packet[3] = 0x23;         // inventory code
       packet[4] = 0x01;         // new inventory mode
       
       //8 bytes here
       //bytes 5 through 12 house the unique ID of the tag (the RFID).      
       byte currByte;
       int shiftCount;
       for(x = 0; x < 8; x++) {
           shiftCount = (7 - x) * 8;            //shift that many places right          
           currByte = (byte) ((tag.rfid >> shiftCount) & 0x000000FF);   //apply bitmask
           packet[5+x] = currByte;       
       }
              
       while(charIndex < 64) {
           //2 more bytes
           packet[13] = (byte) charIndex; //next byte after the tag ID, starting address to read from        
           packet[14] = 0x20;  //Number of blocks to read, might as well be maximum.
           
           //CRC for 2 additional bytes
           int crc = crc16(packet, 15); 
           packet[15] = (byte)(crc & 0xFF);      
           packet[16] = (byte)((crc >> 8) & 0xFF);
           
           try {
               //System.out.println();
               //System.out.println("Reading DATA:");
               //System.out.print(">>> ");
               printAsHex(packet, 17);           

               //begin actual read/write operations. 
               cout.write(packet, 0, 17); //5 + 8 + 3 + 2 + x where x is the number of actual data characters
               cout.flush();
                 
               int count = cin.read(readBuffer);
               if (count > 0) {
                  // System.out.print("<<< ");
                   printAsHex(readBuffer, count);
                }
                System.out.println();
                
                if(readBuffer[3] != 0)
                    throw new RFIDException("Tag not present: " + tag.rfid);
           }
           catch(IOException e) {
               throw(new RFIDException(e));
           }          
           
           //skip the first six bytes
           //length of message contained in first byte
           //skip 2 bytes of checksum at the end
           for(x = 6; x < (int)(readBuffer[0] & 0xFF) - 2; x++) {
               if(((x-1) % 5) != 0) {
                   output[c] = readBuffer[x];
                   c++;
               }                       
           }
                                 
           charIndex += 32;
       }
       
      // System.out.println();
      // System.out.print("Data read: ");
       printAsHex(output, c);
       
       String s = new String(output, 0, c);
       tag.setData(s.substring(0, s.indexOf("\0")));
      // System.out.println(tag.getData());
   }


   private void printAsHex(byte [] data) {
      printAsHex(data, data.length);
   }
   
   private void printAsHex(byte [] data, int len) {
       //for (int i=0; i<len; i++) {
       //System.out.print(Integer.toHexString(0x00FF & data[i]) + " ");
       //}
       //System.out.println();
   }

   
/**
 * Calculates the crc-16 checksum for communications that is
 * appeneded to each instruction.
 */
   private int crc16(byte [] data, int len) {
      int crc = 0x0000FFFF;
      for (int i = 0; i < len; i++)  {
         crc ^= (0x00FF & data[i]);
         for (int j = 0; j < 8; j++) {
            if ((crc & 0x0001) != 0) {
               crc = (crc >> 1) ^ 0x8408;
            } else {
               crc >>= 1;
            }
         }
      }
      return crc;
   }
   

   private void fireEvents(RFIDTag [] thisInventory) {
       Arrays.sort(thisInventory);

       int thisInventoryRunner = 0;
       int lastInventoryRunner = 0;
       while (thisInventoryRunner < thisInventory.length &&
              lastInventoryRunner < lastInventory.length) {
               
           if(lastInventory[lastInventoryRunner].rfid == thisInventory[thisInventoryRunner].rfid) {
               thisInventoryRunner++;
               lastInventoryRunner++;
           }
           else if(lastInventory[lastInventoryRunner].rfid > thisInventory[thisInventoryRunner].rfid) {
               // Throw new event for thisInventory[thisInventoryRunner]
               fireAddEvent(thisInventory[thisInventoryRunner]);
               thisInventoryRunner++;
           }
           else {
               // Throw remove event for lastInventory[lastInventoryRunner]
               fireRemoveEvent(lastInventory[lastInventoryRunner]);
               lastInventoryRunner++;
           }
       }
       for( ; thisInventoryRunner < thisInventory.length; thisInventoryRunner++) {
           // Fire add events
           fireAddEvent(thisInventory[thisInventoryRunner]);
       }
       for( ; lastInventoryRunner < lastInventory.length; lastInventoryRunner++) {
           // Fire remove events
           fireRemoveEvent(lastInventory[lastInventoryRunner]);
       }
       lastInventory = thisInventory;
   }

   
   public void run() {
      try {
         while (!stopped) {
            reset();
            inventory();
            Thread.sleep(250);
         }
      }
      catch (Exception e) { 
          e.printStackTrace();   
      }
   }

   
   private void fireAddEvent(RFIDTag added) {
       for (int i=0; i<listeners.size(); i++) {
           ((RFIDListener)listeners.get(i)).tagAdded(added);
       }
   }

   private void fireRemoveEvent(RFIDTag removed) {
       for (int i=0; i<listeners.size(); i++) {
           ((RFIDListener)listeners.get(i)).tagRemoved(removed);
       }
   }
}
